/*
 * Copyright (c) 2013 CNRS/LAAS
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#ifndef _RMP440_H
#define _RMP440_H

#include <stdint.h>
#include <netinet/in.h>

enum RMP440_INTERFACE_TYPE {
	RMP440_INTERFACE_USB,
	RMP440_INTERFACE_ETHERNET,
	RMP440_INTERFACE_CAN
};

#define RMP_NUM_USB_ETH_BYTES		12

#define RMP_CAN_CMD_ID			0x500
#define RMP_CMD_VEL_WORD_INDEX		0
#define RMP_CMD_YAW_WORD_INDEX		1

#define RMP_CFG_CMD_ID			0x501
#define RMP_CFG_GP_CMD_WORD_INDEX	0

/* configuration commands */
#define RMP_CMD_NONE			0

#define RMP_CMD_SET_MAXIMUM_VELOCITY	1
#define RMP_DEFAULT_MAXIMUM_VELOCITY	2.2352
#define RMP_MAX_VELOCITY		8.047
#define RMP_MIN_VELOCITY		0.0

#define RMP_CMD_SET_MAXIMUM_ACCEL	2
#define RMP_DEFAULT_MAXIMUM_ACCEL	3.923
#define RMP_MAX_ACCEL			7.848
#define RMP_MIN_ACCEL			0.0

#define RMP_CMD_SET_MAXIMUM_DECEL	3
#define RMP_DEFAULT_MAXIMUM_DECEL	3.923
#define RMP_MAX_DECEL			7.848
#define RMP_MIN_DECEL			0.0

#define RMP_CMD_SET_MAXIMUM_DTZ_RATE	4
#define RMP_DEFAULT_MAXIMUM_DTZ_RATE	3.923
#define RMP_MAX_DTZ_RATE		7.848
#define RMP_MIN_DTZ_RATE		0.0

#define RMP_CMD_SET_COASTDOWN_ACCEL	5
#define RMP_DEFAULT_COASTDOWN_ACCEL	1.961
#define RMP_MAX_COASTDOWN_ACCEL		1.961
#define RMP_MIN_COASTDOWN_ACCEL		0.0

#define RMP_CMD_SET_MAXIMUM_TURN_RATE	6
#define RMP_DEFAULT_MAXIMUM_YAW_RATE	3.0
#define RMP_MAX_YAW_RATE		4.5
#define RMP_MIN_YAW_RATE		0

#define RMP_CMD_SET_MAXIMUM_TURN_ACCEL	7
#define RMP_DEFAULT_MAXIMUM_YAW_ACCEL	28.274
#define RMP_MAX_YAW_ACCEL		28.274
#define RMP_MIN_YAW_ACCEL		0.0

#define RMP_CMD_SET_TIRE_DIAMETER	8
#define RMP_X2_TIRE_DIAMETER		0.483616
#define RMP_MAX_TIRE_DIAMETER		1.0
#define RMP_MIN_TIRE_DIAMETER		0.1524

#define RMP_CMD_SET_WHEEL_BASE_LENGTH	9
#define RMP_DEFAULT_WHEEL_BASE_LENGTH	0.5842
#define RMP_MAX_WHEEL_BASE_LENGTH	1.0
#define RMP_MIN_WHEEL_BASE_LENGTH	0.4142

#define RMP_CMD_SET_WHEEL_TRACK_WIDTH	10
#define RMP_DEFAULT_WHEEL_TRACK_WIDTH	0.7112
#define RMP_MAX_WHEEL_TRACK_WIDTH	1.0
#define RMP_MIN_WHEEL_TRACK_WIDTH	0.506476

#define RMP_CMD_SET_TRANSMISSION_RATIO	11
#define RMP_DEFAULT_TRANSMISSION_RATIO	24.2667
#define RMP_MAX_TRANSMISSION_RATIO	200.0
#define RMP_MIN_TRANSMISSION_RATIO	1.0

#define RMP_CMD_SET_INPUT_CONFIG_BITMAP	12
#define RMP_YAW_INPUT_MAPPING		0x01
#define RMP_VEL_INPUT_MAPPING		0x02
#define RMP_AUDIO_SILENCE_REQUEST	0x04
#define RMP_DISABLE_AC_PRESENT_CSI	0x08
#define RMP_DEFAULT_INPUT_CONFIG_BITMAP	0x09

#define RMP_CMD_SET_ETH_IP_ADDRESS	13
#define RMP_DEFAULT_ETH_IP_ADDRESS	0xc0a80040 /* 192.168.0.40 */

#define RMP_CMD_SET_ETH_PORT_NUMBER	14
#define RMP_DEFAULT_ETH_PORT_NUMBER	55

#define RMP_CMD_SET_ETH_SUBNET_MASK	15
#define RMP_DEFAULT_ETH_SUBNET_MASK	0xffffff00 /* 255.255.255.0 */

#define RMP_CMD_SET_ETH_GATEWAY		16
#define RMP_DEFAULT_ETH_GATEWAY		0xc0a80001 /* 192.168.0.1 */

#define RMP_CMD_SET_USER_FB_1_BITMAP	17
#define RMP_DEFAULT_USER_FB_1_BITMAP	0xffffffff
#define RMP_VALID_USER_FB_1_MASK	0xffffffff

#define RMP_CMD_SET_USER_FB_2_BITMAP	18
#define RMP_DEFAULT_USER_FB_2_BITMAP	0xffffffff
#define RMP_VALID_USER_FB_2_MASK	0xffffffff

#define RMP_CMD_SET_USER_FB_3_BITMAP	19
#define RMP_DEFAULT_USER_FB_3_BITMAP	0xffffffff
#define RMP_VALID_USER_FB_3_MASK	0xffffffff

#define RMP_CMD_SET_USER_FB_4_BITMAP	20
#define RMP_DEFAULT_USER_FB_4_BITMAP	0x0
#define RMP_VALID_USER_FB_4_MASK	0x0

#define RMP_CMD_FORCE_CONFIG_FEEDBACK_BITMAPS	30

#define RMP_CMD_SET_AUDIO_COMMAND			31
#define RMP_MOTOR_AUDIO_PLAY_NO_SONG			0
#define RMP_MOTOR_AUDIO_PLAY_POWER_ON_SONG		1
#define RMP_MOTOR_AUDIO_PLAY_POWER_OFF_SONG		2
#define RMP_MOTOR_AUDIO_PLAY_ALARM_SONG			3
#define RMP_MOTOR_AUDIO_PLAY_MODE_UP_SONG		4
#define RMP_MOTOR_AUDIO_PLAY_MODE_DOWN_SONG		5
#define RMP_MOTOR_AUDIO_PLAY_ENTER_ALARM_SONG		6
#define RMP_MOTOR_AUDIO_PLAY_EXIT_ALARM_SONG		7
#define RMP_MOTOR_AUDIO_PLAY_FINAL_SHUTDOWN_SONG	8
#define RMP_MOTOR_AUDIO_PLAY_CORRECT_ISSUE		9
#define RMP_MOTOR_AUDIO_PLAY_ISSUE_CORRECTED		10
#define RMP_MOTOR_AUDIO_PLAY_CORRECT_ISSUE_REPEATING	11
#define RMP_MOTOR_AUDIO_PLAY_BEGINNER_ACK		12
#define RMP_MOTOR_AUDIO_PLAY_EXPERT_ACK			13
#define RMP_MOTOR_AUDIO_ENTER_FOLLOW			14
#define RMP_MOTOR_AUDIO_TEST_SWEEP			15
#define RMP_MOTOR_AUDIO_SIMULATE_MOTOR_NOISE		16

#define RMP_CMD_SET_OPERATIONAL_MODE	32

#define RMP_DISABLE_REQUEST	1
#define RMP_POWERDOWN_REQUEST	2
#define RMP_DTZ_REQEST		3
#define RMP_STANDBY_REQUEST	4
#define RMP_TRACTOR_REQUEST	5
#define RMP_BALANCE_REQUEST	6 /* unused */

#define RMP_CMD_SEND_SP_FAULTLOG	33

#define RMP_FAULTLOG_INITIAL_REQUEST	1
#define RMP_FAULTLOG_NORMAL_REQUEST	0
#define RMP_FAULTLOG_NUM_OF_PACKETS	6
#define RMP_FAULTLOG_PACKET_NUM_OF_WORDS	53

#define RMP_CMD_RESET_INTEGRATORS	34
#define RMP_RESET_LINEAR_POSITION	0x00000001
#define RMP_RESET_RIGHT_FRONT_POSITION	0x00000002
#define RMP_RESET_LEFT_FRONT_POSITION	0x00000004
#define RMP_RESET_RIGHT_REAR_POSITION	0x00000008
#define RMP_RESPONSE_START_CAN_ID	0x502

#define RMP_CMD_RESET_PARAMS_TO_DEFAULT	35

/*----------------------------------------------------------------------*/
/*
 * Fault status words definitions
 */

#define RMP_NO_FAULTS				0x0
#define RMP_ALL_FAULTS				0xffffffff

/*----------------------------------------------------------------------*/

#define CCU_FSW_1_INDEX                      0
#define CCU_FSW_2_INDEX                      1
#define CCU_FSW_3_INDEX                      2
#define CCU_FSW_4_INDEX                      3
#define CCU_MCU_0_FSW_INDEX                  4
#define CCU_MCU_1_FSW_INDEX                  5
#define CCU_MCU_2_FSW_INDEX                  6
#define CCU_MCU_3_FSW_INDEX                  7
#define CCU_FRAME_COUNT_INDEX                8
#define CCU_OPERATIONAL_STATE_INDEX          9
#define CCU_DYNAMIC_RESPONSE_INDEX           10
#define CCU_MIN_PROP_SOC_INDEX               11
#define CCU_AUX_BATT_SOC_INDEX               12
#define CCU_INERTIAL_X_ACC_INDEX             13
#define CCU_INERTIAL_Y_ACC_INDEX             14
#define CCU_INERTIAL_X_RATE_INDEX            15
#define CCU_INERTIAL_Y_RATE_INDEX            16
#define CCU_INERTIAL_Z_RATE_INDEX            17
#define CCU_PSE_PITCH_INDEX                  18
#define CCU_PSE_PITCH_RATE_INDEX             19
#define CCU_PSE_ROLL_INDEX                   20
#define CCU_PSE_ROLL_RATE_INDEX              21
#define CCU_PSE_YAW_RATE_INDEX               22
#define CCU_PSE_DATA_VALID_INDEX             23
#define CCU_YAW_RATE_LIMIT_INDEX             24
#define CCU_VELOCITY_LIMIT_INDEX             25
#define CCU_LINEAR_ACCEL_INDEX               26
#define CCU_LINEAR_VELOCITY_INDEX            27
#define CCU_DIFF_WHEEL_VELOCITY_INDEX        28
#define CCU_RF_WHEEL_VEL_INDEX               29
#define CCU_LF_WHEEL_VEL_INDEX               30
#define CCU_RR_WHEEL_VEL_INDEX               31
#define CCU_LR_WHEEL_VEL_INDEX               32
#define CCU_RF_WHEEL_POS_INDEX               33
#define CCU_LF_WHEEL_POS_INDEX               34
#define CCU_RR_WHEEL_POS_INDEX               35
#define CCU_LR_WHEEL_POS_INDEX               36
#define CCU_LINEAR_POS_INDEX                 37
#define CCU_RF_MOTOR_CURRENT_INDEX           38
#define CCU_LF_MOTOR_CURRENT_INDEX           39
#define CCU_RR_MOTOR_CURRENT_INDEX           40
#define CCU_LR_MOTOR_CURRENT_INDEX           41
#define CCU_MAX_MOTOR_CURRENT_INDEX          42
#define CCU_RF_MOTOR_CURRENT_LIMIT_INDEX     43
#define CCU_LF_MOTOR_CURRENT_LIMIT_INDEX     44
#define CCU_RR_MOTOR_CURRENT_LIMIT_INDEX     45
#define CCU_LR_MOTOR_CURRENT_LIMIT_INDEX     46
#define CCU_MIN_CURRENT_LIMIT_INDEX          47
#define CCU_FPB_1_SOC_INDEX                  48
#define CCU_FPB_2_SOC_INDEX                  49
#define CCU_RPB_1_SOC_INDEX                  50
#define CCU_RPB_2_SOC_INDEX                  51
#define CCU_FPB_1_TEMP_INDEX                 52
#define CCU_FPB_2_TEMP_INDEX                 53
#define CCU_RPB_1_TEMP_INDEX                 54
#define CCU_RPB_2_TEMP_INDEX                 55
#define CCU_VEL_TARGET_INDEX                 56
#define CCU_YAW_TARGET_INDEX                 57
#define CCU_ANGLE_TARGET_INDEX               58
#define CCU_AUX_BATT_VOLTAGE_INDEX           59
#define CCU_AUX_BATT_CURRENT_INDEX           60
#define CCU_AUX_BATT_TEMP_INDEX              61
#define CCU_AUX_BATT_ABB_STATUS_INDEX        62
#define CCU_AUX_BATT_STATUS_INDEX            63
#define CCU_AUX_BATT_FAULTS_INDEX            64
#define CCU_7P2V_BATTERY_VOLTAGE_INDEX       65
#define CCU_SP_SW_BUILD_ID_INDEX             66
#define CCU_UIP_SW_BUILD_ID_INDEX            67
#define CCU_MCU_0_INST_POWER_INDEX           68
#define CCU_MCU_1_INST_POWER_INDEX           69
#define CCU_MCU_2_INST_POWER_INDEX           70
#define CCU_MCU_3_INST_POWER_INDEX           71
#define CCU_MCU_0_TOTAL_ENERGY_INDEX         72
#define CCU_MCU_1_TOTAL_ENERGY_INDEX         73
#define CCU_MCU_2_TOTAL_ENERGY_INDEX         74
#define CCU_MCU_3_TOTAL_ENERGY_INDEX         75
#define CCU_FRAM_VEL_LIMIT_INDEX             76
#define CCU_FRAM_ACCEL_LIMIT_INDEX           77
#define CCU_FRAM_DECEL_LIMIT_INDEX           78
#define CCU_FRAM_MAX_DTZ_DECEL_INDEX         79
#define CCU_FRAM_COASTDOWN_ACCEL_INDEX       80
#define CCU_FRAM_YAW_RATE_LIMIT_INDEX        81
#define CCU_FRAM_YAW_ACCEL_LIMIT_INDEX       82
#define CCU_FRAM_TIRE_DIAMETER_INDEX         83
#define CCU_FRAM_WHEEL_BASE_INDEX            84
#define CCU_FRAM_WHEEL_TRACK_WIDTH_INDEX     85
#define CCU_FRAM_TRANSMISSION_RATIO_INDEX    86
#define CCU_FRAM_CFG_BITMAP_INDEX            87
#define CCU_FRAM_ETH_IP_ADDRESS_INDEX        88
#define CCU_FRAM_ETH_PORT_NUMBER_INDEX       89
#define CCU_FRAM_ETH_SUBNET_MASK_INDEX       90
#define CCU_FRAM_ETH_GATEWAY_INDEX           91
#define CCU_FRAM_USER_FEEDBACK_1_INDEX       92
#define CCU_FRAM_USER_FEEDBACK_2_INDEX       93
#define CCU_FRAM_USER_FEEDBACK_3_INDEX       94
#define CCU_FRAM_USER_FEEDBACK_4_INDEX       95
#define CCU_CHECKSUM_INDEX                   96

struct rmp440_feedback {
	uint32_t fault_status[4];
	uint32_t mcu_fault_status[4];
	float frame_count;
	uint32_t operational_state;
	uint32_t dynamic_response;
	float min_propulsion_batt_soc; /* min. of all batt. state of charge */
	float aux_batt_soc;	       /* aux batt. state of charge */
	float inertial_x_acc;
	float inertial_y_acc;
	float inertial_x_rate;
	float inertial_y_rate;
	float inertial_z_rate;
	float pse_pitch;
	float pse_pitch_rate;
	float pse_roll;
	float pse_roll_rate;
	float pse_yaw_rate;
	uint32_t pse_data_is_valid;
	float yaw_rate_limit;
	float vel_limit;
	float linear_accel;
	float linear_vel;
	float differential_wheel_vel;
	float right_front_vel;
	float left_front_vel;
	float right_rear_vel;

	float left_rear_vel;
	float right_front_pos;
	float left_front_pos;
	float right_rear_pos;
	float left_rear_pos;
	float linear_pos;
	float right_front_current;
	float left_front_current;
	float right_rear_current;
	float left_rear_current;
	float max_motor_current;
	float right_front_current_limit;
	float left_front_current_limit;
	float right_rear_current_limit;
	float left_rear_current_limit;
	float min_motor_current_limit;
	float front_base_batt_1_soc;
	float front_base_batt_2_soc;
	float rear_base_batt_1_soc;
	float rear_base_batt_2_soc;
	float front_base_batt_1_temp;
	float front_base_batt_2_temp;
	float rear_base_batt_1_temp;
	float rear_base_batt_2_temp;
	float vel_target;
	float yaw_rate_target;
	float angle_target;	/* only used on omni platforms */
	float aux_batt_voltage;
	float aux_batt_current;
	float aux_batt_temp;
	uint32_t abb_system_status;
	uint32_t abb_batt_status;

	uint32_t aux_batt_faults;
	float ccu_7p2_battery_voltage;
	uint32_t sp_sw_build_id;
	uint32_t uip_sw_build_id;
	float mcu_inst_power[4];
	float mcu_total_energy[4];
	float fram_vel_limit;
	float fram_accel_limit;
	float fram_decel_limit;
	float fram_dtz_decel_limit;
	float fram_coastdown_decel;
	float fram_yaw_rate_limit;
	float fram_yaw_accel_limit;
	float fram_tire_diameter;
	float fram_wheel_base_length;
	float fram_wheel_track_width;
	float fram_transmission_ratio;
	uint32_t fram_config_bitmap;
	in_addr_t fram_eth_ip_address;
	uint16_t fram_eth_port_number;
	in_addr_t fram_eth_subnet_mask;
	in_addr_t fram_eth_gateway;
	uint32_t user_feedback_bitmap[4];
};

struct rmp440_config {
	uint32_t video_server_online;
	struct in_addr video_server_addr;
	char *video_server_username;
	char *video_server_password;
	char * ocu_controller_type;
	enum RMP440_INTERFACE_TYPE ocu_comm_interface;
	double max_vel;
	double max_accel;
	double max_decel;
	double max_dtz_decel;
	double coastdown_decel;
	double max_yaw_rate;
	double max_yaw_accel;
	double tire_diameter;
	double wheel_base_length;
	double wheel_track_width;
	double gearbox_ratio;
	uint32_t config_bitmap;
	struct in_addr ip_address;
	uint16_t port;
	struct in_addr ip_netmask;
	struct in_addr ip_gateway;
	uint32_t feedback_bitmap[4];
	uint32_t loaded;
};

/* io.c */
extern struct rmp440_io *rmp440Init(enum RMP440_INTERFACE_TYPE, const char *);
extern void rmp440End(struct rmp440_io *);


/* decode.c */
extern struct rmp440_feedback *rmp440FeedbackInit(const char *);
extern const char *rmp440DecodeOperationalState(uint32_t);
extern int rmp440ReceiveAndDecode(const struct rmp440_io *,
    struct rmp440_feedback *);

/* command.c */
extern int rmp440Motion(const struct rmp440_io *, double, double);
extern int rmp440CmdNone(const struct rmp440_io *);
extern int rmp440SetMaximumVelocity(const struct rmp440_io *, double);
extern int rmp440SetMaximumAcceleration(const struct rmp440_io *, double);
extern int rmp440SetMaximumDeceleration(const struct rmp440_io *, double);
extern int rmp440SetMaximumDTZDecelRate(const struct rmp440_io *, double);
extern int rmp440SetCoastDownAccel(const struct rmp440_io *, double);
extern int rmp440SetMaximumTurnRate(const struct rmp440_io *, double);
extern int rmp440SetMaximumTurnAccel(const struct rmp440_io *, double);
extern int rmp440SetTireDiameter(const struct rmp440_io *, double);
extern int rmp440SetWheelBaseLength(const struct rmp440_io *, double);
extern int rmp440SetWheelTrackWidth(const struct rmp440_io *, double);
extern int rmp440SetTransmissionRatio(const struct rmp440_io *, double);
extern int rmp440SetInputConfigBitmap(const struct rmp440_io *, uint32_t);
extern int rmp440SetUserFeedBack(const struct rmp440_io *, uint32_t*);
extern int rmp440SetAudioCommand(const struct rmp440_io *, uint32_t);
extern int rmp440SetInetParams(const struct rmp440_io *,
    uint32_t, uint16_t, uint32_t, uint32_t);
extern int rmp440SetOperationalMode(const struct rmp440_io *, uint32_t);

/* config.c */
struct rmp440_config *rmp440ParseConfigFile(const char *);
void rmp440PrintConfig(const struct rmp440_config *, FILE *);
void rmp440FeedbackFromConfig(struct rmp440_feedback *,
    const struct rmp440_config *);
#endif
